/*
 * sat68kdb: Sega Saturn MC68EC000 Debugger
 * Bart Trzynadlowski
 */

/*
 * driv_sh2.c: Master SH-2 driver
 */

/*
 * Notes:
 * ------
 *
 * The SH-2 will not check for misaligned data access. It is the PC's task to
 * prevent the user from trying to access unaligned data.
 */

#include "defs_sh2.h"
#include "driv_68k.h"


/*
 * void Delay(LONG delay_value);
 *
 * A simple delay function.
 */

void Delay(LONG delay_value)
{
    while (delay_value--)   ;
}


/*
 * BYTE PARTransferByte(BYTE data);
 *
 * Reads data from the PC, and then writes data to the PC.
 */

BYTE PARTransferByte(BYTE data)
{
    volatile BYTE   *flag = (BYTE *) PAR_ADDR_FLAG;
    BYTE            in;

    /*
     * Wait for data flag to become 1
     */

    while (!(*flag & 1))    ;

    /*
     * Read
     */

    in = *((BYTE *) PAR_ADDR_READ);

    /*
     * Write
     */

    *((BYTE *) PAR_ADDR_WRITE) = data;

    return in;
}

/*
 * WORD PARReceiveWord();
 *
 * Calls PARTransferByte 2 times, thus receiving data in big endian format.
 */

WORD PARReceiveWord()
{
    return (PARTransferByte(0) << 8) | PARTransferByte(0);
}

/*
 * void PARSendLong(LONG data);
 *
 * Calls PARTransferByte() 4 times, thus sending data in big endian form.
 */

void PARSendLong(LONG data)
{
    PARTransferByte(data >> 24);
    PARTransferByte((data >> 16) & 0xff);
    PARTransferByte((data >> 8) & 0xff);
    PARTransferByte(data & 0xff);
}

/*
 * LONG PARReceiveLong();
 *
 * Calls PARTransferByte() 4 times, thus receiving data in big endian format.
 */

LONG PARReceiveLong()
{
    LONG    data;

    data = PARTransferByte(0) << 24;
    data |= PARTransferByte(0) << 16;
    data |= PARTransferByte(0) << 8;
    data |= PARTransferByte(0);

    return data;
}


/*
 * void SCSPUploadData(LONG offset, BYTE *data, LONG size);
 *
 * Uploads data to MC68EC000 RAM.
 */

void SCSPUploadData(LONG offset, BYTE *data, LONG size)
{
    LONG    i;

    /*
     * Set up SCSP to allow code to be uploaded
     */

    *((BYTE *) (SCSP_ADDR_CTRL)) = 2;

    /*
     * Upload!
     */

    for (i = 0; i < size; i += 2)
        *((WORD *) (SCSP_ADDR_BASE + offset + i)) = *((WORD *) &data[i]);
}
    
/*
 * void SCSPUploadDriver(BYTE *code, WORD size);
 *
 * Uploads the MC68EC000 driver. The first 10 long words are loaded at offset
 * 0, the rest is loaded at 0x7fe00. Thus, the driver may not exceed (10 * 4)
 * + 0x200 bytes.
 */

void SCSPUploadDriver(BYTE *code, WORD size)
{
    LONG    i, j;

    /*
     * Set up SCSP to allow code to be uploaded
     */

    *((BYTE *) (SCSP_ADDR_CTRL)) = 2;

    /*
     * Clear the memory, first
     */

    for (i = 0; i < 0x80000; i += 2)
        *((WORD *) (SCSP_ADDR_BASE + i)) = 0;

    /*
     * Upload first 10 long words of vector table
     */

    i = 0;
    for (j = 0; j < (10 * 4); j += 2)
    {
        *((WORD *) (SCSP_ADDR_BASE + j)) = *((WORD *) &code[i]);
        i += 2;
    }

    /*
     * Upload the rest to 0x7fe00
     */

    for (j = 0x7fe00; i < size; j += 2)
    {
        *((WORD *) (SCSP_ADDR_BASE + j)) = *((WORD *) &code[i]);
        i += 2;        
    }
}


/*
 * void SMPCCommand(BYTE command);
 *
 * Sends a command to the SMPC. Only Type B commands are currently supported.
 * The operation of a Type B command, as specified by the SMPC manual, is:
 *
 *      - Wait for SF to be reset (to 0, I suppose)
 *      - Set SF (to 1)
 *      - Write command
 *
 * The manual says "read command" instead of "write", so I guess it's looking
 * at things from the SMPC's point of view. I'm not sure how that affects the
 * rest of the operation :/
 */

void SMPCCommand(BYTE command)
{
    volatile BYTE   *sf = (BYTE *) SMPC_ADDR_SF;

    /*
     * Make sure SMPC is not busy
     */

    while (*sf & 0x01)  ;

    /*
     * Send the command
     */

    *sf = 0x01;
    *((BYTE *) SMPC_ADDR_COMREG) = command;
}    


/*
 * void _main();
 *
 * This is where program execution begins and where most of the work is done.
 */

void _main()
{
    volatile WORD   *driv68k_ready = (WORD *) DRIV68K_ADDR_READY;
    volatile WORD   *driv68k_command = (WORD *) DRIV68K_ADDR_COMMAND;
    volatile LONG   *driv68k_data = (LONG *) DRIV68K_ADDR_DATA;
    volatile LONG   *driv68k_data2 = (LONG *) DRIV68K_ADDR_DATA2;
    LONG            i, j, k;

    /*
     * Turn the display off
     */

    *((WORD *) 0x5f80000) = 0;

    /*
     * Turn off the sound CPU
     */

    SMPCCommand(SMPC_CMND_SNDOFF);
    Delay(DELAY_VALUE);

    /*
     * Upload MC68EC000 driver
     */

    SCSPUploadDriver(driv_68k, DRIV68K_SIZE);

    /*
     * Turn on sound CPU
     */

    SMPCCommand(SMPC_CMND_SNDON);
    Delay(DELAY_VALUE);
                       
    /*
     * Initiate communication with the PC:
     *
     *      - PC sends 'M'
     *      - Saturn sends '6'
     *      - PC sends '8'
     *      - Saturn sends 'K'
     */

    while (1)
    {
        if (PARTransferByte('6') != 'M')
            continue;
        if (PARTransferByte('K') != '8')
            continue;
        break;
    }

    /*
     * Now, the PC is going to want to send us a long word telling us the
     * size of the program it's going to send
     */

    i = PARReceiveLong();

    /*
     * Read the program and copy it to SCSP RAM starting at address 0 word by
     * word
     */

    j = 0;
    while (i)
    {
        *((WORD *) (SCSP_ADDR_BASE + j)) = PARReceiveWord();
        j += 2;
        i -= 2;
    }

    /*
     * Tell the 68K it can now begin execution of the loaded program
     */

    *((WORD *) DRIV68K_ADDR_STARTSIG) = 1;
    Delay(DELAY_VALUE);

    /*
     * Main loop: Process commands
     */

    while (1)
    {
        /*
         * Get command and arguments from PC
         */

        i = PARReceiveWord();           /* get command from PC */
        j = PARReceiveLong();           /* get argument from PC */
        k = PARReceiveLong();           /* get second argument from PC */

        /*
         * Process the command. Certain commands are handled by the SH-2, but
         * most are passed to the 68K driver.
         */

        switch (i)
        {
        case CMND_READMEMB:

            k = *((WORD *) (SCSP_ADDR_BASE + (j & ~1)));
            k >>= (j & 1) ? 0 : 8;
            k &= 0xff;
            PARSendLong(k);
            break;

        case CMND_READMEMW:

            PARSendLong(*((WORD *) (SCSP_ADDR_BASE + j)));
            break;
 
        case CMND_READMEML:

            i = (*((WORD *) (SCSP_ADDR_BASE + j)) << 16);
            i |= *((WORD *) (SCSP_ADDR_BASE + j + 2));
            PARSendLong(i);
            break;

        case CMND_WRITEMEMB:

            k &= 0xff;
            i = *((WORD *) (SCSP_ADDR_BASE + (j & ~1)));
            if (j & 1)
                k |= (i & 0xff00);
            else
                k = (i & 0xff) | (k << 8);
            j &= ~1;

            /* next, we write it back as a word */

        case CMND_WRITEMEMW:

            *((WORD *) (SCSP_ADDR_BASE + j)) = (k & 0xffff);
            PARSendLong(0);
            break;

        case CMND_WRITEMEML:

            *((WORD *) (SCSP_ADDR_BASE + j)) = (k >> 16);
            *((WORD *) (SCSP_ADDR_BASE + j + 2)) = k & 0xffff;
            PARSendLong(0);
            break;

        case CMND_RESET:

            while (1)   SMPCCommand(SMPC_CMND_SYSRES);

        default:

            /*
             * Wait until 68K is no longer busy
             */

            while (!(*driv68k_ready))   ;

            /*
             * Send the command and data to the 68K
             */

            *driv68k_command = i;
            *driv68k_data = j;
            *driv68k_data2 = k;

            /*
             * Tell 68K we've finished sending the command and wait
             */

            *driv68k_ready = 0;
            while (!(*driv68k_ready))   ;

            /*
             * Send resultant data to PC
             */

            PARSendLong(*driv68k_data);
            break;
        }
    }
}
